package ru.ifmo.arraylockfreequeue;

import java.util.*;
import java.util.concurrent.atomic.AtomicReferenceArray;

/**
 * Single array-based concurrent queue implementation.
 *
 * @author Iskander Akishev
 */
public class ArrayConcurrentQueue<E> extends AbstractQueue<E> {

	private final int N;
	private final AtomicReferenceArray a;
	private int head, tail;

	private final static Object NULL_UNPROCESSED = new Object();
	private final static Object NULL_PROCESSED = new Object();

	public ArrayConcurrentQueue(int length) {
		this.N = length;
		this.a = new AtomicReferenceArray(length);
		for (int i = 0; i < N; i++) {
			a.set(i, NULL_UNPROCESSED);
		}
		head = tail = 0;
	}

	public int size() {
		while (true) {
			int h = head;
			if (h == N) {
				return 0;
			}
			Object hv = a.get(h);
			if (hv == NULL_PROCESSED) {
				head = ++h;
				continue;
			}

			int t = tail;
			Object tv;
			if (t == N) {
				tv = NULL_UNPROCESSED;
			} else {
				tv = a.get(t);
			}
			if (tv != NULL_UNPROCESSED) {
				tail = t + 1;
				continue;
			}
			return t - h;
		}
	}

	public boolean offer(E e) {
		if (e == null) {
			throw new NullPointerException();
		}
		while (true) {
			int t = tail;
			if (t == N) {
				return false;
			}
			while (a.get(t) != NULL_UNPROCESSED) {
				t++;
				if (t == N) {
					tail = N;
					return false;
				}
			}
			if (a.compareAndSet(t, NULL_UNPROCESSED, e)) {
				tail = t + 1;
				return true;
			}
		}
	}

	public E poll() {
		int h = head;
		if (h == N) {
			return null;
		}
		while (true) {
			Object v = a.get(h);
			if (v == NULL_PROCESSED) {
				head = ++h;
				if (h == N) {
					return null;
				}
			} else if (v == NULL_UNPROCESSED) {
				return null;
			} else {
				if (a.compareAndSet(h, v, NULL_PROCESSED)) {
					head = h + 1;
					return (E) v;
				}
			}
		}
	}

	public E peek() {
		int h = head;
		if (h == N) {
			return null;
		}
		while (true) {
			Object v = a.get(h);
			if (v == NULL_PROCESSED) {
				head = ++h;
				if (h == N) {
					return null;
				}
			} else if (v == NULL_UNPROCESSED) {
				return null;
			} else {
				return (E) v;
			}
		}
	}

	public boolean isExhausted() {
		return (a.get(N - 1) == NULL_PROCESSED);
	}

	public Iterator<E> iterator() {
		throw new UnsupportedOperationException("Iterator is not supported");
	}

}
