package ru.ifmo.arraylockfreequeue;

import java.util.*;
import java.util.concurrent.atomic.AtomicReferenceFieldUpdater;

/**
 * Lock-free concurrent queue implementation, based on unrolled linked list.
 * 
 * @author Iskander Akishev
 */
public class UnrolledListConcurrentQueue<E> extends AbstractQueue<E> implements Queue<E> {

	private final int NODE_SIZE;

	public UnrolledListConcurrentQueue(int nodeSize) {
		this.NODE_SIZE = nodeSize;
	}

	private static final
	    AtomicReferenceFieldUpdater<UnrolledListConcurrentQueue, Node>
	    tailUpdater =
	    AtomicReferenceFieldUpdater.newUpdater
	    (UnrolledListConcurrentQueue.class, Node.class, "tail");
	private static final
	    AtomicReferenceFieldUpdater<UnrolledListConcurrentQueue, Node>
	    headUpdater =
	    AtomicReferenceFieldUpdater.newUpdater
	    (UnrolledListConcurrentQueue.class,  Node.class, "head");
	private static final
	    AtomicReferenceFieldUpdater<UnrolledListConcurrentQueue, Node>
	    unfullUpdater =
	    AtomicReferenceFieldUpdater.newUpdater
	    (UnrolledListConcurrentQueue.class,  Node.class, "unfull");

	private boolean casTail(Node<ArrayConcurrentQueue<E>> cmp, Node<ArrayConcurrentQueue<E>> val) {
	    return tailUpdater.compareAndSet(this, cmp, val);
	}
	private boolean casHead(Node<ArrayConcurrentQueue<E>> cmp, Node<ArrayConcurrentQueue<E>> val) {
	    return headUpdater.compareAndSet(this, cmp, val);
	}
	private boolean casUnfull(Node<ArrayConcurrentQueue<E>> cmp, Node<ArrayConcurrentQueue<E>> val) {
	    return unfullUpdater.compareAndSet(this, cmp, val);
	}

	private volatile Node<ArrayConcurrentQueue<E>> head = new Node<ArrayConcurrentQueue<E>>(null, null);
	private volatile Node<ArrayConcurrentQueue<E>> tail = head;
	private volatile Node<ArrayConcurrentQueue<E>> unfull = head;


	public int size() {
		int count = 0;
		for (Node<ArrayConcurrentQueue<E>> p = getFirst(); p != null; p = p.getNext()) {
			ArrayConcurrentQueue<E> a = p.getItem();
		    if (a != null) {
			    count += a.size();
			    if (count < 0) {
				    return Integer.MAX_VALUE;
			    }
		    }
		}
		return count;
	}

	public boolean offer(E e) {
		if (e == null) throw new NullPointerException();
		while (true) {
			while (true) {
				Node<ArrayConcurrentQueue<E>> u = unfull;
				ArrayConcurrentQueue<E> unfullArray = u.getItem();
				if (unfullArray != null) {
					if (unfullArray.offer(e)) {
						return true;
					}
				}
				Node<ArrayConcurrentQueue<E>> next = u.getNext();
				if (next == null) {
					break;
				}
				casUnfull(u, next);
			}

			// Failed to insert into existing nodes. Need to create new one.
			Node<ArrayConcurrentQueue<E>> n = new Node<ArrayConcurrentQueue<E>>(new ArrayConcurrentQueue<E>(NODE_SIZE), null);
			while (true) {
			    Node<ArrayConcurrentQueue<E>> t = tail;
			    Node<ArrayConcurrentQueue<E>> s = t.getNext();
			    if (t == tail) {
			        if (s == null) {
			            if (t.casNext(s, n)) {
			                casTail(t, n);
			                break;
			            }
			        } else {
			            casTail(t, s);
			        }
			    }
			}
		}
	}

	Node<ArrayConcurrentQueue<E>> getFirst() {
	    while (true) {
	        Node<ArrayConcurrentQueue<E>> h = head;
	        Node<ArrayConcurrentQueue<E>> t = tail;
	        Node<ArrayConcurrentQueue<E>> first = h.getNext();
	        if (h == head) {
	            if (h == t) {
	                if (first == null)
	                    return null;
	                else
	                    casTail(t, first);
	            } else {
	                if (first.getItem() != null)
	                    return first;
	                else // remove deleted node and continue
	                    casUnfull(h, first);
	                    casHead(h, first);
	            }
	        }
	    }
	}

	public E poll() {
		while (true) {
			Node<ArrayConcurrentQueue<E>> h = head;
			Node<ArrayConcurrentQueue<E>> t = tail;
			Node<ArrayConcurrentQueue<E>> first = h.getNext();
			if (h == head) {
			    if (h == t) {
			        if (first == null)
			            return null;
			        else
			            casTail(t, first);
			    } else {
				    ArrayConcurrentQueue<E> a = first.getItem();
				    if (a == null) {
					    casUnfull(h, first);
					    casHead(h, first);
				    } else {
					    E item = a.poll();
					    if (item != null) {
						    return item;
					    } else {
						    if (a.isExhausted()) {
							    casUnfull(h, first);
						        casHead(h, first);
						    } else {
							    return null;
						    }
					    }
				    }
			    }
			}
		}
	}

	public E peek() {
		while (true) {
			Node<ArrayConcurrentQueue<E>> h = head;
			Node<ArrayConcurrentQueue<E>> t = tail;
			Node<ArrayConcurrentQueue<E>> first = h.getNext();
			if (h == head) {
			    if (h == t) {
			        if (first == null)
			            return null;
			        else
			            casTail(t, first);
			    } else {
				    ArrayConcurrentQueue<E> a = first.getItem();
				    if (a == null) {
					    casUnfull(h, first);
					    casHead(h, first);
				    } else {
					    E item = a.peek();
					    if (item != null) {
						    return item;
					    } else {
						    if (a.isExhausted()) {
							    casUnfull(h, first);
						        casHead(h, first);
						    } else {
							    return null;
						    }
					    }
				    }
			    }
			}
		}
	}

	public void clear() {
		head = tail = unfull = new Node<ArrayConcurrentQueue<E>>(null, null);
	}

	public Iterator<E> iterator() {
		throw new UnsupportedOperationException("Iterator is not supported");
	}

}
